import logging
import time
import psutil
from proto_proto_pscr.cb_psc_nta_pb2 import NtaRequest, Flush
from llanta_manager.utils import ScheduledWorkerThread

_logger = logging.getLogger(__name__)


class LlantaPurger(ScheduledWorkerThread):
    """Purge exited/unresponsive/idle llantas.
    This class needs to have references to llanta manager's thread lock and llanta pool.
    Extends the ScheduledWorkerThread class, so we can run it in the background.
    """

    def __init__(
        self,
        lock,
        llanta_pool,
        llanta_unresponsive_timeout_seconds,
        llanta_idle_timeout_seconds,
        llanta_shutdown_wait_seconds,
        purge_interval_seconds,
        zmq_connector,
        zmq_id,
        metrics,
    ):
        self.lock = lock
        self.llanta_pool = llanta_pool
        self.llanta_unresponsive_timeout_seconds = llanta_unresponsive_timeout_seconds
        self.llanta_idle_timeout_seconds = llanta_idle_timeout_seconds
        self.llanta_shutdown_wait_seconds = llanta_shutdown_wait_seconds
        self.purge_interval_seconds = purge_interval_seconds
        self.zmq_connector = zmq_connector
        self.zmq_id = zmq_id
        self.metrics = metrics
        super().__init__(
            job=self.purge_llantas, job_interval_seconds=self.purge_interval_seconds
        )

    def purge_llantas(self, purge_all: bool = False):
        """Purge exited/unresponsive/idle llantas"""
        with self.lock:
            to_purge = []
            if purge_all:
                to_purge = list(self.llanta_pool.keys())
                for llanta_id in to_purge:
                    self.send_shutdown_request(llanta_id)
                _logger.info(
                    f"Waiting {self.llanta_shutdown_wait_seconds} seconds for llantas to shutdown properly."
                )
                time.sleep(self.llanta_shutdown_wait_seconds)
            else:
                for llanta_id, llanta in list(self.llanta_pool.items()):
                    if (
                        not llanta.process.is_running()
                        or llanta.process.status() == psutil.STATUS_ZOMBIE
                    ):
                        # process has exited
                        _logger.info(f"{llanta_id} has exited")
                        to_purge.append(llanta_id)
                        self.metrics.metrics_registry.meter(
                            self.metrics.llanta_purged_exited
                        ).mark()
                    elif (not llanta.ready) and (
                        time.time() - llanta.last_contacted
                    ) >= self.llanta_unresponsive_timeout_seconds:
                        # process seems to be stuck/unresponsive
                        _logger.info(f"{llanta_id} seems to be stuck/unresponsive")
                        to_purge.append(llanta_id)
                        self.metrics.metrics_registry.meter(
                            self.metrics.llanta_purged_unresponsive
                        ).mark()
                    elif llanta.ready:
                        if not llanta.shutdown_requested and (
                            (time.time() - llanta.last_contacted)
                            >= self.llanta_idle_timeout_seconds
                        ):
                            # llanta has been idle for a long time
                            _logger.info(f"{llanta_id} has been idle for a long time")
                            self.send_shutdown_request(llanta_id)
                            self.metrics.metrics_registry.meter(
                                self.metrics.llanta_purged_idle
                            ).mark()
                        elif llanta.shutdown_requested and (
                            (time.time() - llanta.last_contacted)
                            >= self.llanta_shutdown_wait_seconds
                        ):
                            # a shutdown request had already been sent to the llanta
                            _logger.info(
                                f"{llanta_id} has had a shutdown request already sent, terminating process"
                            )
                            to_purge.append(llanta_id)

                    # report llanta process cpu/memory usage
                    if (
                        llanta.process.is_running()
                        and llanta.process.status() != psutil.STATUS_ZOMBIE
                    ):
                        self.metrics.metrics_registry.gauge(
                            f"{self.metrics.llanta_process_cpu}", extra_labels={"llanta_id": llanta_id}
                        ).set_value(llanta.process.cpu_percent())
                        self.metrics.metrics_registry.gauge(
                            f"{self.metrics.llanta_process_memory}", extra_labels={"llanta_id": llanta_id}
                        ).set_value(llanta.process.memory_percent())

            for llanta_id in to_purge:
                _logger.info(f"Purging: {llanta_id}")
                if self.llanta_pool[llanta_id].process.is_running():
                    self.llanta_pool[llanta_id].process.kill()
                del self.llanta_pool[llanta_id]

            self.metrics.metrics_registry.gauge(
                self.metrics.llanta_processes_count
            ).set_value(len(self.llanta_pool))

    def send_shutdown_request(self, llanta_id: bytes):
        """Send a proper shutdown request to the specified llanta through zmq"""
        _logger.info(f"Sending shutdown request to {llanta_id}")
        shutdown_request = NtaRequest(**{"flush": Flush(**{"terminate": True})})
        message = [
            llanta_id,
            b"",
            shutdown_request.SerializeToString(),
        ]
        self.llanta_pool[llanta_id].set_ready(False)
        self.llanta_pool[llanta_id].shutdown_requested = True
        self.zmq_connector.send_to_llanta(message)
