"""Demo for the Llanta manager broker/proxy component with mocked Llantas and NTAs
    - communication between mocked NTA and mocked Llantas through the llanta manager
    - dynamic spawning of llantas
    - mapping between NTA clients and llanta clients
    - purge logic for stuck/failed llantas
    - retry logic on NTA client (simple pirate pattern)
    - alerts publishing from llatnas to KPL workers
"""
import logging
import json
import os
from threading import Lock

from llanta_manager.zmq_connector import ZmqConnector
from llanta_manager.llanta_spawner import LlantaSpawner, MockSpawner
from llanta_manager.llanta_purger import LlantaPurger
from llanta_manager.subnets_monitor import SubnetsMonitor
from llanta_manager.sqs_connector import SqsConnector
from llanta_manager.batch_job_triggerer import BatchJobTriggerer
from llanta_manager.utils import SigHandler, Config
from llanta_manager.metrics import Metrics
from pscr_common.log import setup_logging, clear_logging_handlers


_logger = logging.getLogger(__name__)


class LlantaManager:
    """Main Llanta manager class to handle
    communication and mapping between NTA threads and llantas
    and dynamic spawning of llantas
    """

    # ID in case llanta manager needs to send messages itself
    ZMQ_ID = b"llanta-manager"

    def __init__(self):
        self.config = Config.create()

        clear_logging_handlers()
        setup_logging(
            level=self.config.get("service", "log_level"),
            svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
            env_id=os.environ.get("ENV_ID", "UNKNOWN"),
            svc_id=os.environ.get("SVC_ID", "TODO"),
        )

        # map shards to llantas ("shard_id": Llanta)
        self.llanta_pool = {}
        # use a lock to make purging thread safe
        self.lock = Lock()

        self.nta_endpoint_port = self.config.getint("zmq", "nta_endpoint_port")
        self.llanta_endpoint_host = self.config.get("zmq", "llanta_endpoint_host")
        self.llanta_endpoint_port = self.config.getint("zmq", "llanta_endpoint_port")
        self.nta_alerts_endpoint_host = self.config.get(
            "zmq", "nta_alerts_endpoint_host"
        )
        self.nta_alerts_endpoint_port = self.config.getint(
            "zmq", "nta_alerts_endpoint_port"
        )

        # purge llantas if unresponsive for this long
        llanta_unresponsive_timeout_seconds = self.config.getint(
            "llanta", "unresponsive_timeout_seconds"
        )
        # purge llanta if no traffic for this long
        llanta_idle_timeout_seconds = self.config.getint(
            "llanta", "idle_timeout_seconds"
        )
        llanta_shutdown_wait_seconds = self.config.getint(
            "llanta", "shutdown_wait_seconds"
        )
        purge_interval_seconds = self.config.getint("llanta", "purge_interval_seconds")
        subnets_request_interval_seconds = self.config.getint(
            "llanta", "subnets_request_interval_seconds"
        )
        batch_jobs = (
            json.loads(self.config.get("batch_job", "batch_jobs").strip("/"))
            if self.config.get("batch_job", "batch_jobs")
            else None
        )
        batch_job_sqs_url = self.config.get("batch_job", "sqs_url")
        self.llanta_ready_message = self.config.get_special(
            "llanta", "ready_message"
        ).encode()
        self.healthcheck_request_message = self.config.get_special(
            "manager", "healthcheck_request"
        ).encode()
        self.healthcheck_up_message = self.config.get_special(
            "manager", "healthcheck_response_up"
        ).encode()
        self.healthcheck_down_message = self.config.get_special(
            "manager", "healthcheck_response_down"
        ).encode()
        self.poll_timeout = self.config.getint("zmq", "poll_timeout")

        self.zmq_connector = ZmqConnector(
            self.nta_endpoint_port, self.llanta_endpoint_port, self.poll_timeout
        )
        self.sqs_connector = SqsConnector(batch_job_sqs_url)
        self.metrics = Metrics()
        self.llanta_spawner = (
            LlantaSpawner()
            if not self.config.getboolean("llanta", "mocked")
            else MockSpawner()
        )
        self.llanta_purger = LlantaPurger(
            self.lock,
            self.llanta_pool,
            llanta_unresponsive_timeout_seconds,
            llanta_idle_timeout_seconds,
            llanta_shutdown_wait_seconds,
            purge_interval_seconds,
            self.zmq_connector,
            self.ZMQ_ID,
            self.metrics,
        )
        self.subnets_monitor = SubnetsMonitor(
            self.llanta_pool,
            self.zmq_connector,
            subnets_request_interval_seconds,
        )
        self.batch_job_triggerers = []
        if batch_jobs:
            self.batch_job_triggerers = [
                BatchJobTriggerer(
                    self.llanta_pool,
                    detector_id,
                    int(batch_job_interval_seconds),
                    self.sqs_connector,
                )
                for detector_id, batch_job_interval_seconds in batch_jobs.items()
            ]

    @staticmethod
    def _get_llanta_id(nta_id: bytes) -> bytes:
        """Mapping of NTA client ids to llanta instance ids
        (nta_id will probably be shard id)"""
        return f"llanta-{nta_id.decode()}".encode()

    def _healthcheck(self, nta_id: bytes):
        """Perform healthcheck and respond with up or down
        (for now only respond with up)
        """
        message = [nta_id, b"", self.healthcheck_up_message]
        self.zmq_connector.send_to_nta(message)

    def process_llanta_message(self, llanta_id: bytes, message: bytes):
        """Process messages coming from llantas"""
        if not message:
            _logger.error(f"Message from llanta {llanta_id} is None")
            return

        if llanta_id not in self.llanta_pool:
            _logger.warning(f"Receiving message from unknown llanta: {llanta_id}")
            self.metrics.metrics_registry.meter(
                self.metrics.llanta_failed_lookups
            ).mark()
            return

        # mark llanta as ready
        self.llanta_pool[llanta_id].set_ready(True)
        nta_id = self.llanta_pool[llanta_id].nta_id

        if self.llanta_pool[llanta_id].shutdown_requested:
            # llanta is responding after a shutdown request was sent (probably to the shutdown request)
            _logger.info(
                f"Llanta {llanta_id} is responding after a shutdown request was sent"
            )
        elif message == self.llanta_ready_message:
            # llanta process has started, process the last message now that llanta is ready
            self.process_nta_message(nta_id, self.llanta_pool[llanta_id].last_message)
        elif self.llanta_pool[llanta_id].subnets_requested:
            self.subnets_monitor.process_active_subnets_response(llanta_id, message)
        else:
            # forward message to NTA
            message = [nta_id, b"", message]
            _logger.debug(f"Forwarding to NTA: {message}")
            self.zmq_connector.send_to_nta(message)

            # check if we need to update the active subnets
            self.subnets_monitor.monitor_llanta(llanta_id)

    def process_nta_message(self, nta_id: bytes, message: bytes):
        """Process messages coming from NTAs"""
        if not message:
            _logger.error(f"Message from NTA {nta_id} is None")
            return

        llanta_id = self._get_llanta_id(nta_id)

        if message == self.healthcheck_request_message:
            self._healthcheck(nta_id)
        elif llanta_id not in self.llanta_pool:
            self.llanta_pool[llanta_id] = self.llanta_spawner.spawn_llanta(
                llanta_id,
                nta_id,
                self.llanta_endpoint_host,
                self.llanta_endpoint_port,
                self.nta_alerts_endpoint_host,
                self.nta_alerts_endpoint_port,
                message,
            )
        else:
            if self.llanta_pool[llanta_id].shutdown_requested:
                _logger.warning(
                    f"NTA {nta_id} is trying to send messages to llanta {llanta_id} after a shutdown request, ignoring..."
                )
            elif self.llanta_pool[llanta_id].ready:
                # forward to corresponding llanta instance
                message = [llanta_id, b"", message]
                _logger.debug(f"Forwarding to llanta: {message}")
                self.zmq_connector.send_to_llanta(message)
                # mark lanta as busy
                self.llanta_pool[llanta_id].set_ready(False)
                self.metrics.metrics_registry.meter(
                    f"{self.metrics.llanta_events_rate}", extra_labels={"llanta_id": llanta_id}
                ).mark()
            else:
                _logger.warning(
                    f"NTA {nta_id} is retrying messages before llanta is ready"
                )
                # buffer message to be sent when llanta is ready
                self.llanta_pool[llanta_id].last_message = message

    def start(self):
        """Listen on both sockets"""
        self.zmq_connector.bind()
        self.llanta_purger.start()
        if self.batch_job_triggerers:
            for triggerer in self.batch_job_triggerers:
                triggerer.start()
        else:
            _logger.warning(
                "No batch jobs config specified. Llanta manager will not trigger jobs for active subnets."
            )

        _logger.info("Listening")
        sig_handler = SigHandler()
        ids_without_active_subnets = set()
        while True:
            got_llanta_messages, got_nta_messages = self.zmq_connector.poll_sockets()

            if got_llanta_messages:
                _logger.debug("llanta message received")
                llanta_id, message = self.zmq_connector.read_llanta_message()
                with self.lock:
                    self.process_llanta_message(llanta_id, message)

            if got_nta_messages:
                _logger.debug("nta message received")
                nta_id, message = self.zmq_connector.read_nta_message()
                with self.lock:
                    self.process_nta_message(nta_id, message)

            if sig_handler.interrupted:
                if len(ids_without_active_subnets) == len(self.llanta_pool):
                    break
                self.subnets_monitor.subnets_request_interval_seconds = 10
                for llanta_id in list(self.llanta_pool.keys()):
                    if len(self.llanta_pool[llanta_id].active_subnets) != 0:
                        self.subnets_monitor.monitor_llanta(llanta_id)
                    else:
                        ids_without_active_subnets.add(llanta_id)

        _logger.info("Exiting...")
        self.llanta_purger.purge_llantas(purge_all=True)
        self.zmq_connector.close_sockets()


def main():
    manager = LlantaManager()
    manager.config.print_config_props()
    manager.start()


if __name__ == "__main__":
    main()
