import sys
import zmq
import random
from time import sleep
from proto_proto_pscr.cb_psc_nta_pb2 import (
    NtaRequest,
    NtaResponse,
    NtaAlert,
    LicenseInfo,
)
import uuid


class LlantaMock:
    """Dummy zmq client that reads and responds to messages,
    simulating work, alerts and random crashes
    """

    # signal to llanta manager, that given llanta instance is ready
    LLANTA_READY = "\x01"

    def __init__(self, llanta_id):
        self.llanta_id = llanta_id

    def _create_sockets(self, ll_manager_endpoint, alerts_endpoint):
        self.ctx = zmq.Context()
        self.ll_manager_socket = self.ctx.socket(zmq.REQ)
        self.ll_manager_socket.setsockopt(zmq.IDENTITY, self.llanta_id)
        self.ll_manager_socket.connect(ll_manager_endpoint)

        self.alerts_socket = self.ctx.socket(zmq.REQ)
        self.alerts_socket.connect(alerts_endpoint)

        # tell llanta manager we're ready
        self.ll_manager_socket.send(self.LLANTA_READY.encode())

    def _close_sockets(self):
        self.ll_manager_socket.setsockopt(zmq.LINGER, 0)
        self.ll_manager_socket.close()

        self.alerts_socket.setsockopt(zmq.LINGER, 0)
        self.alerts_socket.close()

    def _generate_response_proto(self, reply):
        response_args = {
            "response_code": NtaResponse.NTA_RESPONSE_CODE_OK
        }
        if reply:
            response_args.update(reply)
        response_proto = NtaResponse(**response_args)

        return response_proto

    def _generate_alert_proto(self, alert):
        subnet_id = LicenseInfo(
            **{"user_id": 123, "access_key_id": 123, "subkey_id": 123}
        )

        alert_args = {
            "event_references": ["event1", "event2"],
            "subnet_id": subnet_id,
            "alert_id": f"alert-{uuid.uuid4()}",
            "alert_hash": "hash",
            "confidence_score": random.randint(0, 100),
            "detector": "detector",
            "detector_type": NtaAlert.NETFLOW,
        }
        alert_proto = NtaAlert(**alert_args)

        return alert_proto

    def decode_message(self, message):
        request = NtaRequest()
        request.ParseFromString(message)
        return request

    def send_reply(self, reply, additional_fields=None):
        response_proto = self._generate_response_proto(additional_fields)
        reply[-1] = response_proto.SerializeToString()
        self.ll_manager_socket.send_multipart(reply)

    def process_events_message(self, message):
        print(f"Received events in {self.llanta_id}")
        self.send_reply(message)

    def process_configure_message(self, message):
        print(f"Configured {self.llanta_id}")
        self.send_reply(message)

    def process_flush_message(self, message):
        print(f"Flushed {self.llanta_id}")
        self.send_reply(message)

    def process_query_status_message(self, message):
        print(f"Queried status of {self.llanta_id}")
        additional_fields = {
            "license_infos": [
                LicenseInfo(**{"user_id": 123, "access_key_id": 123, "subkey_id": 123}),
                LicenseInfo(**{"user_id": 234, "access_key_id": 234, "subkey_id": 234}),
            ]
        }
        self.send_reply(message, additional_fields)

    def process_shutdown_message(self, message):
        print(f"Shutting down {self.llanta_id}")
        self.send_reply(message)
        self._close_sockets()
        sys.exit(0)

    def message_switch(self, message):
        message[-1] = self.decode_message(message[-1])
        message_type = message[-1].WhichOneof("request")

        if message_type == "nta_net_events":
            self.process_events_message(message)
        elif message_type == "configure":
            self.process_configure_message(message)
        elif message_type == "flush":
            if message[-1].flush.terminate:
                self.process_shutdown_message(message)
            else:
                self.process_flush_message(message)
        elif message_type == "query_status_request":
            self.process_query_status_message(message)
        else:
            print("Unknown message!")

    def publish_alert(self, alert):
        alert_proto = self._generate_alert_proto(alert)
        self.alerts_socket.send(alert_proto.SerializeToString())
        # wait for ack from NTA KPL
        self.alerts_socket.recv()

    def start(self, ll_manager_endpoint, alerts_endpoint):
        print(f"Staring Llanta {self.llanta_id}")
        self._create_sockets(ll_manager_endpoint, alerts_endpoint)

        loops = 0
        while True:
            msg = self.ll_manager_socket.recv_multipart()
            if not msg:
                break
            print(f"llanta {self.llanta_id} received: {msg}")

            if random.randint(0, 10) >= 7:
                # random alert!
                self.publish_alert(f"Alert from {self.llanta_id}".encode())

            # simulate failure
            # if loops >= 5 and random.randint(0, 10) >= 5:
            #     if random.randint(0, 10) >= 5:
            #         # crashed
            #         raise Exception("Failed!")
            #     else:
            #         # stuck
            #         while True:
            #             sleep(2)

            # simulate work being done
            # sleep(0.2)
            self.message_switch(msg)
            loops += 1
